from rest_framework import serializers


class LocationSerializer(serializers.Serializer):
    name = serializers.CharField(max_length=255)
    lat  = serializers.FloatField()
    lng  = serializers.FloatField()



from django.conf import settings
from typing import Sequence
from tsp_wrapper import BROKER_URL

class Settings:
    """
    Shadow Django's settings with a little logic
    """
    @property
    def TSP_BROKER_URL(self) -> str:
        return getattr(settings, "TSP_BROKER_URL", BROKER_URL)

    @property
    def TSP_HOOK_URL(self) -> str:
        return getattr(settings, "TSP_HOOK_URL", "http://localhost:8000/api/tsp/hook/")

conf = Settings()


from channels.layers import get_channel_layer
from tsp_wrapper import broker, INBOUND_QUEUE, setup
from asgiref.sync import async_to_sync
import secrets 
from kombu import Connection
from django_tsp.conf import conf


channel_layer = get_channel_layer()

def notify_event(event:dict) -> None:
    """
    call sockets
    """
    async_to_sync(channel_layer.group_send)('tsp', {'type': 'add_event', 'event': event})


def publish_tsp_sloution(data):
    data["type"] = "answer"
    notify_event(data)

def create_tsp_problem(data):
    data["type"] = "question"
    data["id"] = secrets.token_hex(5)  # TODO: store the task ids 
    notify_event(data)
    broker.send_data(connection=broker.get_broker(), message=data, routing_key=INBOUND_QUEUE)


def prepare_broker():
    """
        changing the broker url with settings.py variables
        and connecting to the broker
        and setting up the middlewares
    """
    broker.set_broker(Connection(conf.TSP_BROKER_URL))
    setup()


from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework import status, serializers
from drf_spectacular.utils import extend_schema
from .services import create_tsp_problem, publish_tsp_sloution
from .serializers import LocationSerializer 


class SolverApi(APIView):
    """
        Accept the tsp problem instance and 
        broadcast question to websocket 
    """
    class InputSolverSerializer(serializers.Serializer):
        locations = LocationSerializer(many=True)
        num_vehicles = serializers.FloatField(default=1)
        depot = serializers.FloatField(default=0)

    @extend_schema(request=InputSolverSerializer)
    def post(self, request):
        serializer = self.InputSolverSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            create_tsp_problem(
                request.data
            )
        except Exception as ex:
            return Response(
                    f"Database Error {ex}",
                    status=status.HTTP_400_BAD_REQUEST
                    )
        return Response(status=status.HTTP_204_NO_CONTENT)

class WebhookApi(APIView):
    """
        Recive the sloution response and notify websocket clients
        TODO: make this endpoint private or using other portocol for handeling the sloution
    """
    class InputSerializer(serializers.Serializer):
        locations = serializers.ListSerializer(child=serializers.CharField())
        id = serializers.CharField()

    @extend_schema(request=InputSerializer)
    def post(self, request):
        serializer = self.InputSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            publish_tsp_sloution(
                request.data
            )
        except Exception as ex:
            return Response(
                    f"Database Error {ex}",
                    status=status.HTTP_400_BAD_REQUEST
                    )
        return Response(status=status.HTTP_204_NO_CONTENT)


from django.core.management.base import BaseCommand
from tsp_wrapper import broker, tsp_solver_task
from django_tsp.services import prepare_broker

class Command(BaseCommand):
    help = 'This command will activate tsp_wrapper solver task'

    def handle(self, *args, **kwargs):
        prepare_broker()
        self.stdout.write(self.style.SUCCESS('Starting the Bridge.'))
        worker = broker.Worker(connection=broker.get_broker(), handler=tsp_solver_task, type="inbound")
        worker.run()


from django.core.management.base import BaseCommand
from tsp_wrapper import broker, OUTBOUND_QUEUE
from django_tsp.services import prepare_broker
import requests
from django_tsp.conf import conf


class Command(BaseCommand):
    help = 'Consume the tsp sloution and bordcast it to the /ws/tsp socket'

    def handle(self, *args, **kwargs):
        prepare_broker()
        self.stdout.write(self.style.SUCCESS('Starting the Reciver.'))
        worker = broker.Worker(connection=broker.get_broker(), handler=self.socket_tsp, type=OUTBOUND_QUEUE)
        worker.run()

    def socket_tsp(self, res):
        """
        our costume task for notifing the sloution to the web socket client
        """
        response = requests.post(conf.TSP_HOOK_URL, json = res)
        self.stdout.write(self.style.NOTICE(f'socket_tsp task response --> {response}'))